/*
** NetParse
**
** Parse network traffic for matching patterns and return strings
**
** Author:		Julian Anastasov <ja@ssi.bg>, January 2001
**
**			Released under the GPL version 2
**
*/

#define NETPARSE_VERSION	"0.6"
#define NP_DEBUG
#define	WITH_SSL



#include <sys/types.h>
#include <linux/types.h>
#include <sys/times.h>
#include <sys/param.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <ctype.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <linux/errqueue.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <resolv.h>

#ifdef	WITH_SSL
#include <ssl.h>
#include <err.h>
#include <e_os.h>
#endif

#include "list.h"
#include "timer.h"
#include "npclient.h"
#include "bregex.h"

#ifdef  sun
int gethostname(char *name,int namelen);
#endif

#define	MAX_TIMEOUT	(30*60)
#define	MAX_GET_SIZE	(1*1024*1024)

#define state_t		struct struct_state
#define conn_t		struct struct_conn

state_t {
	state_t		*next;
	int		refcnt;
	int		type;
	int		timeout;
	int		escape;
	int		flags;
	int		size;
	char		*str;
	int		str_size;
	char		*fnm;
};

conn_t {
	char		*host;
	unsigned	port;
	long		sid;
	int		ttl;
	state_t		*states;
	state_t		*cur_state;
	list_t		timer;
	int		fd;
	int		flags;		/* CONN_xxx */
	int		status;		/* Connection status */
	struct timeval	time_start;
	int		time_resolve;	/* ms */
	int		time_connect;	/* ms */
	char		*buf;
	int		bufsize;
	int		rpos;
	int		wpos;
	int		sendpos;
	FILE		*fp;		/* Used from GETTOFILE */
	unsigned long	regex_matches;
	int		regex_pos;
	char		*result;
	int		result_size;
	int		result_pos;
#ifdef	WITH_SSL
	SSL		*ssl;
	int		ssl_code;
#else
	void		*ssl;	/* Let "if (conn_t.ssl)" to work */
#endif
};

#define	STATE_UNUSED		0
#define	STATE_CONNECTING	1
#define	STATE_CONNECTED		2

/* state_t.flags: */
#define	FLAG_ONLY_FAILED	0x0001
#define	FLAG_CASEMATCH		0x0002
#define	FLAG_HOLD_SSL		0x0004
#define	FLAG_PRINT_PROTO	0x0008
#define	FLAG_PRINT_SERVICE	0x0010
#define	FLAG_PRINT_SID		0x0020
#define	FLAG_PRINT_STATUS	0x0040
#define	FLAG_PRINT_TIME_RESOLVE	0x0080
#define	FLAG_PRINT_TIME_CONNECT	0x0100

#define	CONN_CAN_READ		0x01
#define	CONN_CAN_WRITE		0x02
#define	CONN_IN_SHUTDOWN	0x04
#define	CONN_EOF		0x08
#define	CONN_SSL		0x10
#define	CONN_UDP		0x20

enum {
	TYPE_WAIT,
	TYPE_GET,
	TYPE_GETTIME,
	TYPE_GETTOFILE,
	TYPE_SKIP,
	TYPE_SEND,
	TYPE_SHUTDOWN,
	TYPE_SSL,
	TYPE_CLOSE,
};

char *debug_cmd_arr[] = {
	[TYPE_WAIT]		= "WAIT",
	[TYPE_GET]		= "GET",
	[TYPE_GETTIME]		= "GETTIME",
	[TYPE_GETTOFILE]	= "GETTOFILE",
	[TYPE_SKIP]		= "SKIP",
	[TYPE_SEND]		= "SEND",
	[TYPE_SHUTDOWN]		= "SHUTDOWN",
	[TYPE_SSL]		= "SSL",
	[TYPE_CLOSE]		= "CLOSE",
};


/* Can't be more than FD_SETSIZE */
#define	NUM_SOCKETS		1000

#ifdef	WITH_SSL
static int with_ssl = 1;
#else
static int with_ssl = 0;
#endif

#ifdef	WITH_SSL
static SSL_CTX			*ctx_ssl;
#endif

static int			use_ssl = 0;
static int			parse_only = 0;	/* Parse the input data only */
static int			have_one_request = 0;
static int			have_file_requests = 0;
static int			def_timeout = 5;
static int			def_nsocks = 20;
static int			min_sock = FD_SETSIZE;
static int			max_sock = 0;
static int			ns = 0;
static int			nready;
static int			socks = 0;
static int			state[FD_SETSIZE];
static struct sockaddr_in	*sock_ad[FD_SETSIZE];
static conn_t			*conns[FD_SETSIZE];
static state_t			*def_states = 0;
static char			*def_command = 0;
static char			*one_request = 0;
static char			*buffer = 0;
static int			bufsize = 4096;
static int			max_get_size = 500*1024;
static timetable_t		tt;
static struct sockaddr_in	ad;
static fd_set			rfs, readfs, wfs, writefs;
static FILE			*fpin = 0;
static FILE			*fpout;
static volatile int		flag_sig_alrm = 0;
static int			res_retrans = 0;
static int			res_retry = 0;

//extern struct __res_state _res;

/*
** debug levels:
** 0 = quiet
** 1 = fatal errors
** 2 = non-fatal errors
** 3 = debugging
** 4 = more debugging
*/


#ifdef	NP_DEBUG
static int np_debug_level = 1;
#define	debug(l,f...) {if (l <= np_debug_level) { \
	int _np_err = errno; fprintf(stderr, ## f); errno = _np_err;}}
#define	mem_error() debug(1, "%s:%d: Unable to allocate memory.\n", \
			__FUNCTION__, __LINE__)
#else
#define	debug(l,f...)	/**/
#define	mem_error()	/**/
#endif	/* !NP_DEBUG */


static char *debug_cmd_str(int type)
{
	if (type < 0 || type >= sizeof(debug_cmd_arr)/sizeof(debug_cmd_arr[0])) {
		return "?";
	}
	return debug_cmd_arr[type];
}

static char *debug_sid(conn_t *cp)
{
static char buf[1024];

	sprintf(buf, "SID%ld:%s:%.900s:%u",
		cp->sid,
		(cp->flags & CONN_SSL) ? "SSL":
		((cp->flags & CONN_UDP) ? "UDP" : "TCP"),
		cp->host, cp->port);
	return buf;
}

static void sig_alrm(int sig)
{
	flag_sig_alrm = 1;
}

static state_t *alloc_state(void)
{
state_t *p;

	p = (state_t *) calloc(1, sizeof(*p));
	if (!p) return 0;
	return p;
}

static void free_state(state_t *p)
{
	if (--p->refcnt > 0) return;
	if (p->str) free(p->str);
	if (p->fnm) free(p->fnm);
	if (p->next) free_state(p->next);
	free(p);
}

static conn_t *alloc_conn(void)
{
conn_t *p;

	p = (conn_t *) calloc(1, sizeof(*p));
	if (!p) return 0;
	list_init(&p->timer);
	p->bufsize = 1024;
	return p;
}

static void free_conn(conn_t *p)
{
	if (p->host) free(p->host);
	if (p->states) free_state(p->states);
	if (p->result) free(p->result);
	if (p->buf) free(p->buf);
	if (p->fp) fclose(p->fp);
	if (!list_empty(&p->timer)) list_remove(&p->timer);
	free(p);
}

static int gettime_diff(struct timeval *tv1, struct timeval *tv2)
{
struct timeval tv;
int td, td2;

	if (!tv2) {
		tv2 = &tv;
		if (gettimeofday(tv2, 0) < 0) {
			return 0;
		}
	}
	td = tv2->tv_sec - tv1->tv_sec;
	td2 = tv2->tv_usec - tv1->tv_usec;
	if (td2 < 0) {
		if (td <= 0) return 0;
		td = (td - 1) * 1000 + (1000000 + td2) / 1000;
	} else {
		td = td * 1000 + td2 / 1000;
	}
	return td;
}

static int start_connection(int s)
{
struct hostent *h;
struct in_addr *ia;
int rc1, on;
conn_t *cp = conns[s];
char *host = cp->host;
unsigned port = cp->port;

	memset((char *) &ad,0,sizeof(ad));
	ad.sin_family = AF_INET;
	ad.sin_port = htons(port);
	ad.sin_addr.s_addr = inet_addr(host);
	if (gettimeofday(&cp->time_start, 0) < 0) {
		debug(1, "gettimeofday: %s\n", strerror(errno));
		return -1;
	}
	if (ad.sin_addr.s_addr == -1) {
		h = gethostbyname(host);
		cp->time_resolve = gettime_diff(&cp->time_start, 0);
		if (!h) {
			debug(1, "Unknown host: %s\n", host);
			return -2;
		}
		ia = (struct in_addr *) *h->h_addr_list;
		ad.sin_addr.s_addr = ia->s_addr;
	}

	rc1 = fcntl(s,F_GETFL,0);
	if (rc1 >= 0) {
		rc1 |= O_NONBLOCK;
		rc1 = fcntl(s,F_SETFL,rc1);
		if (rc1 < 0) {
			debug(1, "SETFL[%d](%s:%u): %s\n",
				s, host, port, strerror(errno));
		}
	} else {
		debug(1, "GETFL[%d](%s:%u): %s\n",
			s, host, port, strerror(errno));
	}

	if (rc1 < 0) return -1;

	if (cp->flags & CONN_UDP) {
		on = IP_PMTUDISC_DO;
		rc1 = setsockopt(s, SOL_IP, IP_MTU_DISCOVER, &on, sizeof(on));
		if (rc1 < 0) {
			debug(1, "MTU_DISCOVER[%d](%s:%u): %s\n",
				s, host, port, strerror(errno));
		}
		on = 1;
		rc1 = setsockopt(s, SOL_IP, IP_RECVERR, &on, sizeof(on));
		if (rc1 < 0) {
			debug(1, "RECVERR[%d](%s:%u): %s\n",
				s, host, port, strerror(errno));
		}
	}
	if (cp->ttl) {
		on = cp->ttl;
		rc1 = setsockopt(s, SOL_IP, IP_TTL, &on, sizeof(on));
		if (rc1 < 0) {
			debug(1, "TTL[%d](%s:%u): %s\n",
				s, host, port, strerror(errno));
			return -1;
		}
	}

	/* Start connection in background	*/

	cp->time_connect = cp->time_resolve;

	do {
		rc1 = connect(s, (struct sockaddr *) &ad, sizeof(ad));
	} while (rc1 < 0 && EINTR == errno);
	if (rc1 >= 0) {

		debug(3, "%s: Connection to %s:%u established, fd=%d\n",
			__FUNCTION__, host, port, s);

		/* Connected	*/
		return 1;
	} else if (EINPROGRESS != errno) {
		debug(2, "connect1(%s:%d,%d): %s\n",
			host, port, s, strerror(errno));
		/* Host is DOWN, may be	*/
		return -1;
	}
	/* EINPROGRESS	*/
	debug(3, "%s: Connection to %s:%u in progress, fd=%d\n",
		__FUNCTION__, host, port, s);
	return 0;
}

static void close_sock(int i)
{
	debug(3, "%s: closing fd=%d\n", __FUNCTION__, i);
	if (sock_ad[i]) {
		free(sock_ad[i]);
		sock_ad[i] = 0;
	}
	if (FD_ISSET(i, &wfs)) {
		/* TODO: don't dec nready for some platforms */
		nready --;
		FD_CLR(i, &wfs);
	}
	if (FD_ISSET(i, &rfs)) {
		/* TODO: don't dec nready for some platforms */
		nready --;
		FD_CLR(i, &rfs);
	}
	if (conns[i]) {
		if (!list_empty(&conns[i]->timer))
			list_remove_init(&conns[i]->timer);
		if (conns[i]->fp) {
			fclose(conns[i]->fp);
			conns[i]->fp = 0;
		}
		if (conns[i]->buf) {
			free(conns[i]->buf);
			conns[i]->buf = 0;
		}
#ifdef	WITH_SSL
		if (conns[i]->ssl) {
			SSL_free(conns[i]->ssl);
			conns[i]->ssl = 0;
		}
#endif
	}
	FD_CLR(i,&readfs);
	FD_CLR(i,&writefs);
	state[i] = STATE_UNUSED;
	close(i);
	ns --;
}

static void connect_finished(int i)
{
conn_t *cp = conns[i];

	cp->time_connect = gettime_diff(&cp->time_start, 0);

	debug(3, "%s: Connection to %s established, fd=%d\n",
		__FUNCTION__, debug_sid(cp), i);
	state[i] = STATE_CONNECTED;
	cp->flags |= CONN_CAN_WRITE;
	free(sock_ad[i]);
	sock_ad[i] = 0;
	FD_CLR(i,&readfs);
	FD_CLR(i,&writefs);
}

static int finish_connecting(int i)
{
int rc1;
conn_t *cp = conns[i];

	rc1 = connect(i, (struct sockaddr *) sock_ad[i], sizeof(ad));
	if ((rc1 >= 0) || (EISCONN == errno)) {
		rc1 = 0;
		connect_finished(i);
	}
	else {
		debug(2, "connect2(%s:%u,%d): %s\n",
			conns[i]->host, conns[i]->port, i, strerror(errno));
		cp->time_connect = gettime_diff(&cp->time_start, 0);
		close_sock(i);
	}
	cp->status = rc1;
	return rc1;
}

static int conn_next_timeout(conn_t *cp)
{
int timeout = 0;

	if (cp->cur_state) {
		timeout = cp->cur_state->timeout;
	} else {
		/* May be the connection was in progress */
		if (list_empty(&cp->timer)) {
			timeout = def_timeout;
			if (cp->states && cp->states->timeout) {
				timeout = cp->states->timeout;
			}
		}
	}
	if (timeout) debug(4, "%s: selecting timeout=%d for %s, fd=%d\n",
		__FUNCTION__, timeout, debug_sid(cp), cp->fd);
	return timeout;
}

/*
** Return:
** -1: error
** 0: continue
** 1: closed
*/

static int sock_recverr(int i)
{
int sta;
struct msghdr m;
struct cmsghdr *c;
struct sock_extended_err *e;
struct sockaddr_in msg_name;
struct iovec msg_iov;
struct in_addr ina;
char buf[2048];
char cmsg[1024];
conn_t *cp = conns[i];

	m.msg_name		= &msg_name;
	m.msg_namelen		= sizeof(msg_name);
	m.msg_iov		= &msg_iov;
	m.msg_iovlen		= 1;
	m.msg_flags		= 0;
	m.msg_control		= cmsg;
	m.msg_controllen	= sizeof(cmsg);
	msg_iov.iov_base	= buf;
	msg_iov.iov_len		= sizeof(buf);
	sta = recvmsg(i, &m, MSG_ERRQUEUE);
	if (sta < 0) {
		if (EAGAIN == errno) {
			return 0;
		}
		debug(1, "%s: recvmsg failed for %s: %s\n",
			__FUNCTION__, debug_sid(cp), strerror(errno));
		return -1;
	}
	debug(3, "%s: recvmsg() for %s returns %d\n",
		__FUNCTION__, debug_sid(cp), sta);
	e = 0;
	for (c = CMSG_FIRSTHDR(&m); c; c = CMSG_NXTHDR(&m, c)) {
		if (c->cmsg_level != SOL_IP) continue;
		if (c->cmsg_type == IP_RECVERR) {
			e = (void *) CMSG_DATA(c);
			continue;
		}
		debug(3, "%s: cmsg_type %d not handled\n",
			__FUNCTION__, c->cmsg_type);
	}
	if (!e) return 0;
	ina.s_addr = e->ee_info;
	if (SO_EE_ORIGIN_LOCAL == e->ee_origin) {
		debug(1, "%s: recvmsg for %s returns local error: %s, info=%s\n",
			__FUNCTION__, debug_sid(cp),
			strerror(e->ee_errno), inet_ntoa(ina));
		goto err;
	}
	if (SO_EE_ORIGIN_ICMP == e->ee_origin) {
		debug(1, "%s: recvmsg for %s returns ICMP error (type=%d, code=%d): %s, info=%s\n",
			__FUNCTION__, debug_sid(cp),
			e->ee_type, e->ee_code,
			strerror(e->ee_errno), inet_ntoa(ina));
		goto err;
	}
	debug(1, "%s: Unhandled origin: %d\n", __FUNCTION__, e->ee_origin);
err:
	close_sock(i);
	cp->status = -1;
	return 1;
}

#ifdef	WITH_SSL

static int flag_ssl_initialized = 0;

static int init_ssl_data(void)
{
	SSL_load_error_strings();
	SSLeay_add_ssl_algorithms();
	//SSLv3_client_method ?
	ctx_ssl = SSL_CTX_new(SSLv23_client_method());
	return 1;
}

static inline int set_ssl_code(conn_t *cp, int ret, char *call)
{
int err;

#ifdef	NP_DEBUG
	cp->ssl_code = err = SSL_get_error(cp->ssl, ret);
	if (np_debug_level < 2) return err;
	switch (err) {
	case SSL_ERROR_NONE:
		debug(3, "%s: %s succeeded for %s\n",
			__FUNCTION__, call, debug_sid(cp));
		return -1;
	case SSL_ERROR_SSL:
		debug(2, "%s: %s failed for %s, ERROR_SSL\n",
			__FUNCTION__, call, debug_sid(cp));
		break;
	case SSL_ERROR_WANT_READ:
		FD_SET(cp->fd, &readfs);
		debug(2, "%s: %s failed for %s, WANT_READ\n",
			__FUNCTION__, call, debug_sid(cp));
		break;
	case SSL_ERROR_WANT_WRITE:
		FD_SET(cp->fd, &writefs);
		debug(2, "%s: %s failed for %s, WANT_WRITE\n",
			__FUNCTION__, call, debug_sid(cp));
		break;
	case SSL_ERROR_WANT_X509_LOOKUP:
		debug(2, "%s: %s failed for %s, WANT_X509_LOOKUP\n",
			__FUNCTION__, call, debug_sid(cp));
		break;
	case SSL_ERROR_SYSCALL:
		debug(2, "%s: %s failed for %s, SYSCALL: %s\n",
			__FUNCTION__, call, debug_sid(cp),
			strerror(get_last_socket_error()));
		break;
	case SSL_ERROR_ZERO_RETURN:
		debug(2, "%s: %s failed for %s, ZERO_RETURN\n",
			__FUNCTION__, call, debug_sid(cp));
		break;
	case SSL_ERROR_WANT_CONNECT:
		debug(2, "%s: %s failed for %s, WANT_CONNECT\n",
			__FUNCTION__, call, debug_sid(cp));
		break;
	default:
		debug(2, "%s: %s failed for %s, unknwon code=%d\n",
			__FUNCTION__, call, debug_sid(cp), err);
		break;
	}
	return err;
#endif
}

#endif

static int new_ssl_connection(int i)
{
conn_t *cp = conns[i];
#ifdef	WITH_SSL
SSL *ssl;
int sta, err;

	if (!flag_ssl_initialized) {
		if (!init_ssl_data()) goto out;
		flag_ssl_initialized = 1;
	}
	/* Create new SSL structure */
	if (!(cp->ssl = SSL_new(ctx_ssl))) {
		char buf[256+1];

		ERR_error_string(ERR_get_error(), buf);
		debug(1, "%s: SSL_new failed for %s: %s\n",
			__FUNCTION__, debug_sid(cp), buf);
		goto out;
	}

	ssl = cp->ssl;
	SSL_set_fd(ssl, cp->fd);
	SSL_set_connect_state(ssl);

connect_again:
	err = SSL_connect(ssl);
	set_ssl_code(cp, err, "SSL_connect");
	switch(cp->ssl_code) {
	case SSL_ERROR_NONE:
		break;
	case SSL_ERROR_WANT_WRITE:
		cp->flags &= ~CONN_CAN_WRITE;
		FD_SET(i, &writefs);
		debug(3, "%s: SSL_connect: no write space for %s, fd=%d\n",
			__FUNCTION__, debug_sid(cp), i);
		break;
	case SSL_ERROR_WANT_READ:
		cp->flags &= ~CONN_CAN_READ;
		FD_SET(i, &readfs);
		break;
	case SSL_ERROR_SYSCALL:
		sta = get_last_socket_error();
		if (EINTR == sta) {
			goto connect_again;
		}
		debug(3, "%s: SSL_connect for %s, fd=%d, returns: %s\n",
			__FUNCTION__, debug_sid(cp), i, strerror(sta));
		close_sock(i);
		conns[i]->status = -1;
		return 0;
	case SSL_ERROR_SSL: {
		char buf[256+1];

		ERR_error_string(ERR_get_error(), buf);
		debug(1, "%s: SSL_connect failed for %s: %s\n",
			__FUNCTION__, debug_sid(cp), buf);
		}
		/* continue */
	default:
		debug(1, "%s: code=%d for %s\n",
			__FUNCTION__, cp->ssl_code, debug_sid(cp));
		close_sock(i);
		conns[i]->status = -1;
		return 0;
	}
	return 1;
out:
#else
	/* Unreachable point */
#endif
	close_sock(i);
	cp->status = -1;
	return 0;
}

static int next_state(int i)
{
int timeout;
state_t *old_sp, *sp;
conn_t *cp = conns[i];

	timeout = conn_next_timeout(conns[i]);
	old_sp = cp->cur_state;
	if (old_sp) {
		cp->cur_state = old_sp->next;
	} else {
		cp->cur_state = cp->states;
	}
	if (!(sp = cp->cur_state)) {
		/* No more states */
		close_sock(i);
		cp->status = 0;
		return 0;
	}
	if (timeout) {
		timer_mod(&tt, &cp->timer, timeout);
	}
	/* Initialize some variables */
	cp->sendpos = 0;
	cp->regex_matches = 0;
	/* Now set the rw bits according to the command */
	switch (sp->type) {
	case TYPE_SEND:
		if (!(cp->flags & CONN_CAN_WRITE)) FD_SET(i, &writefs);
		FD_CLR(i, &readfs);
		break;
	default:
		if (!(cp->flags & CONN_CAN_READ)) FD_SET(i, &readfs);
		break;
	case TYPE_CLOSE:
		FD_CLR(i, &readfs);
		FD_CLR(i, &writefs);
		break;
	}
	debug(3, "%s: state changed to type %d, timeout=%d\n",
		__FUNCTION__, sp->type, timeout);
	if (!old_sp && (cp->flags & CONN_SSL) && !(sp->flags & FLAG_HOLD_SSL)) {
		if (!new_ssl_connection(i)) return 0;
	}
	return 1;
}

static int handle_states(int i)
{
conn_t *cp = conns[i];
state_t *sp;
int type;
int sz, rsz, sz0, rsz0, sta, sz2;
int force_read = 0;

	debug(3, "%s: handling states for %s, fd=%d\n",
		__FUNCTION__, debug_sid(cp), i);

	loop:

	sp = cp->cur_state;
	type = sp->type;

	if (cp->flags & CONN_CAN_WRITE
#ifdef	WITH_SSL
		|| (cp->ssl && cp->flags & CONN_CAN_READ &&
			SSL_want_read(cp->ssl))
#endif
		) {
#ifdef	WITH_SSL
		debug(4, "%s: %s: can write because: can_write=%d, want_read=%d\n",
			__FUNCTION__, debug_sid(cp),
			(cp->flags & CONN_CAN_WRITE) != 0,
			(cp->ssl && cp->flags & CONN_CAN_READ &&
			SSL_want_read(cp->ssl)));
#endif

		/* Do we need to write something ? */
		if (TYPE_SEND == type) {
			if (cp->flags & CONN_IN_SHUTDOWN) {
				/* write() after shutdown() is error */
				debug(1, "%s: SEND after SHUTDOWN for %s, fd=%d\n",
					__FUNCTION__, debug_sid(cp), i);
				close_sock(i);
				cp->status = -1;
				return 0;
			}
			/* I don't listen for read and write events */
			FD_CLR(i, &readfs);
			FD_CLR(i, &writefs);
#ifdef	WITH_SSL
			if (cp->ssl) {
				while (cp->sendpos < sp->str_size) {
					sta = SSL_write(cp->ssl,
						sp->str + cp->sendpos,
						sp->str_size - cp->sendpos);
					set_ssl_code(cp, sta, "SSL_write");
					switch(cp->ssl_code) {
					case SSL_ERROR_NONE:
						debug(3, "%s: for %s SSL written %d bytes at pos %d/%d\n",
							__FUNCTION__, debug_sid(cp), sta,
							cp->sendpos, sp->str_size);
						cp->sendpos += sta;
						break;
					case SSL_ERROR_WANT_WRITE:
						cp->flags &= ~CONN_CAN_WRITE;
						FD_SET(i, &writefs);
						debug(3, "%s: no write space at pos %d for %s, fd=%d\n",
							__FUNCTION__, cp->sendpos, debug_sid(cp), i);
						return 0;
					case SSL_ERROR_WANT_READ:
						debug(2, "%s: want_read while writing at pos %d for %s, fd=%d\n",
							__FUNCTION__, cp->sendpos, debug_sid(cp), i);
						cp->flags &= ~CONN_CAN_READ;
						FD_SET(i, &readfs);
						return 0;
					case SSL_ERROR_SYSCALL:
						sta = get_last_socket_error();
						if (EINTR == sta) {
							break;
						}
						debug(3, "%s: SSL_write for %s, fd=%d, returns: %s\n",
							__FUNCTION__, debug_sid(cp), i, strerror(sta));
						close_sock(i);
						cp->status = -1;
						return 0;
					case SSL_ERROR_SSL: {
						char buf[256+1];

						ERR_error_string(ERR_get_error(), buf);
						debug(1, "%s: SSL_write failed for %s: %s\n",
							__FUNCTION__, debug_sid(cp), buf);
						}
						/* continue */
					default:
						debug(1, "%s: code=%d for %s\n",
							__FUNCTION__, cp->ssl_code, debug_sid(cp));
						close_sock(i);
						cp->status = -1;
						return 0;
					}
				}
			} else
#endif
			while (cp->sendpos < sp->str_size) {
				do {
					sta = write(i, sp->str + cp->sendpos,
						sp->str_size - cp->sendpos);
				} while (sta < 0 && EINTR == errno);
				if (sta >= 0) {
					debug(3, "%s: for %s written %d bytes at pos %d/%d\n",
						__FUNCTION__, debug_sid(cp), sta,
						cp->sendpos, sp->str_size);
					cp->sendpos += sta;
				} else {
					if (EAGAIN == errno ||
						EWOULDBLOCK == errno) {
						cp->flags &= ~CONN_CAN_WRITE;
						FD_SET(i, &writefs);
						debug(3, "%s: no write space at pos %d for %s, fd=%d\n",
							__FUNCTION__, cp->sendpos, debug_sid(cp), i);
						return 0;
					} else {
						debug(1, "%s: write(%d,): %s\n",
							__FUNCTION__, i, strerror(errno));
						if (CONN_UDP & cp->flags) {
							sta = sock_recverr(i);
							if (sta > 0) return 0;
							if (!sta) continue;
						}
						close_sock(i);
						cp->status = -1;
						return 0;
					}
				}
			}
			debug(3, "%s: SEND completed for %s, %d bytes\n",
				__FUNCTION__, debug_sid(cp), cp->sendpos);
			if (next_state(i)) goto loop;
			return 0;
		}
	}

	switch (type) {
	case TYPE_SSL:
		debug(4, "%s: Handling command %s\n",
			__FUNCTION__, debug_cmd_str(type));
		if (cp->flags & CONN_IN_SHUTDOWN) {
			/* "shutdown ssl"! Why? */
			debug(1, "%s: SSL after SHUTDOWN for %s, fd=%d\n",
				__FUNCTION__, debug_sid(cp), i);
			close_sock(i);
			cp->status = -1;
			return 0;
		}
		if (cp->flags & CONN_UDP) {
			debug(1, "%s: SSL for UDP? %s, fd=%d\n",
				__FUNCTION__, debug_sid(cp), i);
			close_sock(i);
			cp->status = -1;
			return 0;
		}
		if (!cp->ssl) {
			if (!(cp->flags & CONN_SSL)) {
				debug(1, "%s: Will not start SSL to non-SSL service %s\n",
					__FUNCTION__, debug_sid(cp));
				/* Continue with the next states */
			} else if (!new_ssl_connection(i)) return 0;
		}
		/* SSL is already started */
		if (next_state(i)) goto loop;
		return 0;
	case TYPE_GETTIME:
		debug(4, "%s: Handling command %s\n",
			__FUNCTION__, debug_cmd_str(type));
		if (cp->result_pos + 30 > cp->result_size) {
			int newsz;

			newsz = cp->result_pos + 30;
			if (newsz < 1024) newsz += 128;
			else if (newsz < 65536) newsz += 2048;
			else newsz += 65536;
			debug(4, "%s: reallocating buffer for results, %d bytes\n",
				__FUNCTION__, newsz);
			cp->result_size = newsz;
			if (!(cp->result = realloc(cp->result,
				cp->result_size+1))) {
				return -1;
			}
		}
		if (cp->result_pos) {
			cp->result[cp->result_pos++] = ' ';
		}
		cp->result_pos += sprintf(cp->result+cp->result_pos,
			"%d", gettime_diff(&cp->time_start, 0));
		debug(3, "%s: GETTIME completed\n", __FUNCTION__);
		if (next_state(i)) goto loop;
		return 0;
	case TYPE_SHUTDOWN:
		debug(4, "%s: Handling command %s\n",
			__FUNCTION__, debug_cmd_str(type));
		if (!(cp->flags & CONN_IN_SHUTDOWN)) {
			/* What really means shutdown for SSL ? */
			if (!cp->ssl && !(CONN_UDP & cp->flags)) {
				do {
					sta = shutdown(i, 1);
				} while (sta < 0 && EINTR == errno);
			}
			cp->flags |= CONN_IN_SHUTDOWN;
		} else {
			debug(2, "2nd SHUTDOWN for %s\n",
				debug_sid(cp));
		}
		if (next_state(i)) goto loop;
		return 0;
	case TYPE_CLOSE:
		debug(3, "%s: CLOSE for %s, fd=%d\n",
			__FUNCTION__, debug_sid(cp), i);
		close_sock(i);
		cp->status = 0;
		return 0;
	}

	if (!cp->buf) {
		debug(3, "%s: Allocating connection input buffer for %s, fd=%d\n",
			__FUNCTION__, debug_sid(cp), i);
		if (!(cp->buf = (void *) malloc(cp->bufsize+1))) {
			return -1;
		}
	}

	FD_SET(i, &readfs);
	FD_CLR(i, &writefs);

loop_read:
	if (cp->rpos >= cp->wpos || force_read) {
		if (cp->flags & CONN_EOF) {
			goto l1;
		}
		/* We need more data in buffer */
		if (!(cp->flags & CONN_CAN_READ)) {
			debug(3, "%s: %s, fd=%d needs more data at pos %d/%d\n",
				__FUNCTION__, debug_sid(cp), i,
				cp->rpos, cp->wpos);
			FD_SET(i, &readfs);
			return 0;
		}

		/* Increase the free buffer space if possible */

		if (cp->rpos) {
			if (cp->wpos - cp->rpos < 128 ||
				(cp->bufsize-cp->wpos < 64) || force_read) {
				if (cp->wpos != cp->rpos) {
					memcpy(cp->buf, cp->buf+cp->rpos,
						cp->wpos-cp->rpos);
				}
				cp->wpos -= cp->rpos;
				cp->rpos = 0;
			}
		}

#ifdef	WITH_SSL
		if (cp->ssl) {
			for (; cp->wpos < cp->bufsize;) {
				sta = SSL_read(cp->ssl,
					cp->buf + cp->wpos,
					cp->bufsize - cp->wpos);
				set_ssl_code(cp, sta, "SSL_read");
				switch(cp->ssl_code) {
				case SSL_ERROR_NONE:
					debug(3, "%s: for %s SSL read %d bytes at bufpos %d\n",
						__FUNCTION__, debug_sid(cp),
						sta, cp->wpos);
					cp->wpos += sta;
					break;
				case SSL_ERROR_WANT_WRITE:
					cp->flags &= ~CONN_CAN_WRITE;
					FD_SET(i, &writefs);
					/* Only write event please */
					FD_CLR(i, &readfs);
					debug(3, "%s: SSL_read: no write space for %s, fd=%d\n",
						__FUNCTION__, debug_sid(cp), i);
					return 0;
				case SSL_ERROR_WANT_READ:
					cp->flags &= ~CONN_CAN_READ;
					if (cp->rpos < cp->wpos) {
						goto end_read;
					}
					FD_SET(i, &readfs);
					return 0;
				case SSL_ERROR_ZERO_RETURN:
					/* Do we come here? */
					cp->flags |= CONN_EOF;
					debug(3, "%s: EOF for %s, fd=%d\n",
						__FUNCTION__, debug_sid(cp), i);
					FD_CLR(i, &readfs);
					goto end_read;
				case SSL_ERROR_SYSCALL:
					sta = get_last_socket_error();
					if (EINTR == sta) {
						break;
					}
					if (!sta) {
						/* Oh, what is ZERO_RETURN for? */
						cp->flags |= CONN_EOF;
						debug(3, "%s: EOF for %s, fd=%d\n",
							__FUNCTION__, debug_sid(cp), i);
						FD_CLR(i, &readfs);
						goto end_read;
					}
					debug(3, "%s: SSL_read for %s, fd=%d, returns: %s\n",
						__FUNCTION__, debug_sid(cp), i, strerror(sta));
					close_sock(i);
					cp->status = -1;
					return 0;
				case SSL_ERROR_SSL: {
					char buf[256+1];

					ERR_error_string(ERR_get_error(), buf);
					debug(1, "%s: SSL_read failed for %s: %s\n",
						__FUNCTION__, debug_sid(cp), buf);
					}
					/* continue */
				default:
					debug(1, "%s: code=%d for %s\n",
						__FUNCTION__, cp->ssl_code, debug_sid(cp));
					close_sock(i);
					cp->status = -1;
					return 0;
				}
			}
			end_read:
		} else
#endif
		for (; cp->wpos < cp->bufsize;) {
			/* HOLD_SSL ? */
			sz = (cp->flags & CONN_SSL) ?
				1 : cp->bufsize-cp->wpos;
			do {
				sta = read(i, cp->buf+cp->wpos,
					cp->bufsize-cp->wpos);
			} while (sta < 0 && EINTR == errno);
			if (sta < 0) {
				if (EAGAIN == errno) {
					cp->flags &= ~CONN_CAN_READ;
					if (cp->rpos < cp->wpos) {
						break;
					}
					FD_SET(i, &readfs);
					return 0;
				}
				debug(1, "%s: read() for %s returns: %s\n",
					__FUNCTION__, debug_sid(cp),
					strerror(errno));
				if (CONN_UDP & cp->flags) {
					sta = sock_recverr(i);
					if (sta > 0) return 0;
					if (!sta) continue;
				}
				close_sock(i);
				cp->status = -1;
				return 0;
			} else if (!sta) {
				cp->flags |= CONN_EOF;
				debug(3, "%s: EOF for %s, fd=%d\n",
					__FUNCTION__, debug_sid(cp), i);
				FD_CLR(i, &readfs);
				break;
			} else {
				debug(3, "%s: for %s read %d bytes at bufpos %d\n",
					__FUNCTION__, debug_sid(cp),
					sta, cp->wpos);
				cp->wpos += sta;
				/* Before entering SSL read byte by byte */
				if (cp->flags & CONN_SSL) break;
				if (cp->flags & CONN_UDP) {
					/* The default status on timeout is -1 */
					cp->status = -1;
				}
			}
		}
	}

l1:
	debug(4, "%s: have %d bytes for %s\n",
		__FUNCTION__, cp->wpos-cp->rpos, debug_sid(cp));
	/* rpos < wpos && eof=0 */
	/* rpos < wpos && eof=1 */
	/* rpos = wpos && eof=1 */
	force_read = 0;

	/* Match the regular expression */

	if (!sp->str_size) {
		if (TYPE_WAIT == type) {	/* wait eof */
			cp->rpos = cp->wpos = 0;
			if (cp->flags & CONN_EOF) {
				debug(3, "%s: WAIT: MATCH, EOF detected for %s, fd=%d\n",
					__FUNCTION__, debug_sid(cp), i);
				if (next_state(i)) goto loop;
				return 0;
			}
			debug(4, "%s: WAIT: still waiting for EOF\n",
				__FUNCTION__);
			goto loop;
		}
		if (TYPE_SKIP == type) {	/* match eof */
			if (cp->rpos == cp->wpos && cp->flags & CONN_EOF) {
				debug(3, "%s: SKIP: EOF detected for %s, fd=%d\n",
					__FUNCTION__, debug_sid(cp), i);
				if (next_state(i)) goto loop;
				return 0;
			}
			debug(3, "%s: SKIP: no EOF at current pos for %s, fd=%d\n",
				__FUNCTION__, debug_sid(cp), i);
			close_sock(i);
			cp->status = -1;
			return 0;
		}
		debug(1, "%s:%d: Unhandled condition for fd=%d, type=%d\n",
			__FUNCTION__, __LINE__, i, type);
		/* GET ? */
		return -1;
	}
	rsz0 = rsz = sp->str_size - cp->regex_pos;
	sz0 = sz = cp->wpos-cp->rpos;
	debug(5, "%s: before bregex: input=[%.30s], regex=[%.30s]\n",
		__FUNCTION__, cp->buf+cp->rpos, sp->str+cp->regex_pos);
	sta = bregex_match(cp->buf+cp->rpos, &sz, sp->str+cp->regex_pos, &rsz,
		&cp->regex_matches, sp->escape, !(FLAG_CASEMATCH & sp->flags));
	if ((!sta || -2 == sta) && cp->flags & CONN_EOF &&
		cp->rpos < cp->wpos && !sz) {
		/* Match EOF to regex if it is not yet matched */
		debug(4, "%s: sending explicitly EOF to bregex\n",
			__FUNCTION__);
		sta = bregex_match("", &sz, sp->str+cp->regex_pos+rsz0-rsz,
			&rsz, &cp->regex_matches, sp->escape,
			!(FLAG_CASEMATCH & sp->flags));
	}
	debug(4, "%s: regex for %s returns %d(%s), matched input=[%.*s] from %d bytes, rest %d bytes, matched regex [%.*s] from %d bytes, rest %d bytes\n",
		__FUNCTION__, debug_sid(cp), sta,
		(sta>0)?"OK":(!sta?"NO MATCH":((-1==sta)?
		"REGEX ERROR":"NEED MORE DATA")),
		sz0-sz, cp->buf+cp->rpos, sz0, sz,
		rsz0-rsz, sp->str+cp->regex_pos, rsz0, rsz);
	if (-1 == sta) return -1;
	switch (type) {
	case TYPE_SKIP:
		if (sta > 0) {	/* regex is matched */
			debug(3, "%s: SKIP regex matched\n", __FUNCTION__);
			cp->rpos += (sz0-sz);
			if (next_state(i)) goto loop;
			return 0;
		} else {
			/*
			** We have:
			** sta=0: no match
			** EOF: no match
			** sz=sz0 with full buffer: no progress. Yes, we
			** have max atom size that must be present in the
			** buffer
			*/
			if (!sta || cp->flags & CONN_EOF ||
				(sz == sz0 &&
				cp->wpos - cp->rpos >= cp->bufsize)) {
				debug(3, "%s: SKIP regex check failed for %d bytes in buffer\n",
					__FUNCTION__, cp->wpos - cp->rpos);
				close_sock(i);
				cp->status = -1;
				return 0;
			}
			/* We need more data in buffer, sta=-2 */
			debug(4, "%s: SKIP: skipped %d bytes from input but needs more data\n",
				__FUNCTION__, sz0-sz);
			force_read = 1;
			cp->rpos += (sz0-sz);
			cp->regex_pos += (rsz0-rsz);
			goto loop;
		}
	case TYPE_WAIT:
		if (sta > 0) {
			debug(3, "%s: WAIT regex matched\n", __FUNCTION__);
			if (next_state(i)) goto loop;
			return 0;
		} else {
			if (!sta || cp->wpos - cp->rpos >= (cp->bufsize>>1)) {
				cp->rpos ++;
				if (!sta) {
					debug(4, "%s: WAIT: skipping one char from input because the text is not matched\n",
					__FUNCTION__);
				} else {
					debug(4, "%s: WAIT: skipping one char from input because wpos-rpos=%d >= buf space=%d, no progress\n",
					__FUNCTION__, cp->wpos-cp->rpos, cp->bufsize>>1);
				}
			} else {
				if (cp->flags & CONN_EOF) {
					debug(3, "%s: WAIT failed at EOF, need more data, data in buffer: %d bytes\n",
						__FUNCTION__, cp->wpos - cp->rpos);
					close_sock(i);
					cp->status = -1;
					return 0;
				}
				debug(4, "%s: WAIT: waiting for more data\n",
					__FUNCTION__);
			}
			cp->regex_matches = 0;
			if (!sta && cp->wpos > cp->rpos) goto l1;
			force_read = 1;
			goto loop_read;
		}
	case TYPE_GET:
	case TYPE_GETTOFILE:
		sz2 = sz0 - sz;
		if (sz2 > 0 || (!sz2 && cp->flags & CONN_EOF &&
			sta > 0 && !cp->sendpos)) {
			int newsz;

			/*
			** sz2>0: don't use quotes
			** else: we must print "" because one argument
			** for GET must be printed (when GET matches 0 bytes)
			*/

			if (TYPE_GETTOFILE == type) {
				if (!cp->fp) {
					char buf[1024];
					char *fnm;

					if (cp->sid) {
						sprintf(buf, "%.1000s%ld",
							sp->fnm, cp->sid);
						fnm = buf;
					} else {
						fnm = sp->fnm;
					}
					cp->fp = fopen(fnm, "wb");
					if (!cp->fp) {
						debug(1, "%s: Unable to create output file %s\n",
							__FUNCTION__, fnm);
						close_sock(i);
						cp->status = -1;
						return 0;
					}
				}
				if (sz2 > 0 && (sz2 != fwrite(cp->buf+cp->rpos,
					1, sz2, cp->fp) || ferror(cp->fp))) {
					debug(1, "%s: Write error for out file\n",
						__FUNCTION__);
					close_sock(i);
					cp->status = -1;
					return 0;
				}
				goto l_get1;
			}

			/*
			** Each input char results in up to 3 chars (%XX)
			** if sz2 is 0 we need space for ""
			*/
			if (cp->result_pos + sz2*3 + 3 > cp->result_size) {
				newsz = cp->result_pos + sz2*3;
				if (newsz > max_get_size) {
					debug(1, "%s: too long buffer for %s: %d bytes > %d\n",
						__FUNCTION__, debug_sid(cp),
						newsz, max_get_size);
					close_sock(i);
					cp->status = -1;
					return 0;
				}
				if (newsz < 1024) newsz += 128;
				else if (newsz < 65536) newsz += 2048;
				else newsz += 65536;
				debug(4, "%s: reallocating buffer for results, %d bytes\n",
					__FUNCTION__, newsz);
				cp->result_size = newsz;
				if (!(cp->result = realloc(cp->result,
					cp->result_size+1))) {
					return -1;
				}
			}
			if (!cp->sendpos && cp->result_pos) {
				cp->result[cp->result_pos++] = ' ';
			}
			cp->result_pos += np_store_quoted_string(
				cp->result+cp->result_pos, cp->buf+cp->rpos,
				sz2, 1, (sz2 > 0)?0:'"');

			l_get1:
			cp->sendpos = 1;
			cp->rpos += sz2;
		}
		if (sta > 0) {
			if (TYPE_GETTOFILE == type) {
				if (cp->fp) {
					fclose(cp->fp);
					cp->fp = 0;
				}
			}
			debug(3, "%s: %s completed\n",
				__FUNCTION__, debug_cmd_str(type));
			if (next_state(i)) goto loop;
			return 0;
		}
		if (!sta || cp->flags & CONN_EOF ||
			(cp->wpos - cp->rpos >= cp->bufsize)) {
			debug(3, "%s: %s regex check failed\n",
				__FUNCTION__, debug_cmd_str(type));
			close_sock(i);
			cp->status = -1;
			return 0;
		}
		/* sta=-2 */
		cp->regex_pos += (rsz0-rsz);
		/* Same state, read more data */
		debug(4, "%s: %s: need more data, regex_pos=%d\n",
			__FUNCTION__, debug_cmd_str(type), cp->regex_pos);
		return 0;
	}
	return -1;
}

static int handle_event(int i)
{
int sta;

	if (STATE_CONNECTED == state[i]) {
		return handle_states(i);
	}
	if (STATE_CONNECTING == state[i]) {
		sta = finish_connecting(i);
		if (sta < 0 || !(conns[i]->flags &
			(CONN_CAN_READ|CONN_CAN_WRITE)))
			return 0;
		if (!next_state(i)) return 0;
		return handle_states(i);
	}
	debug(1, "%s: Unhandled state=%d for fd=%d\n",
		__FUNCTION__, state[i], i);

	close_sock(i);
	conns[i]->status = -1;
	return 0;
}

/*
** Analyze the commands for the parser
** If states=0 => parse only
** Returns:
** 0 - ok
** 1 - no data
** 2 - invalid data
** 3 - internal error (nomem)
*/

static int parse_states(char *s, state_t **states)
{
state_t *sp, *slist, **slp;
char *cp, *str, c;
int sz, sz2, sz3, sz4;
unsigned u;
int flags = 0;
int retval, type, timeout;
char escape = '!';

	if (states) *states = 0;
	cp = s + strspn(s, " \t");
	if (!*cp || '\n' == *cp || '\r' == *cp) return 1;

	type = -1;
	timeout = 0;
	slist = 0;
	slp = &slist;
	retval = 2;
	for (;;) {
		sp = 0;
		switch (*cp & 0xDF) {
		case 'W':
			if (!strncasecmp(cp, "wait", 4)) {
				cp += 4;
				type = TYPE_WAIT;
				goto arg_str;
			}
			goto fail;
		case 'S':
			if (!strncasecmp(cp, "send", 4)) {
				cp += 4;
				type = TYPE_SEND;
				goto arg_str;
			}
			if (!strncasecmp(cp, "skip", 4)) {
				cp += 4;
				type = TYPE_SKIP;
				goto arg_str;
			}
			if (!strncasecmp(cp, "service", 7)) {
				cp += 7;
				flags |= FLAG_PRINT_SERVICE;
				break;
			}
			if (!strncasecmp(cp, "status", 6)) {
				cp += 6;
				flags |= FLAG_PRINT_STATUS;
				break;
			}
			if (!strncasecmp(cp, "sid", 3)) {
				cp += 3;
				flags |= FLAG_PRINT_SID;
				break;
			}
			if (!strncasecmp(cp, "shutdown", 8)) {
				cp += 8;
				sp = alloc_state();
				if (!sp) goto error;
				*slp = sp;
				slp = &sp->next;
				sp->type = TYPE_SHUTDOWN;
				goto fill_common;
			}
			if (!strncasecmp(cp, "ssl", 3)) {
				cp += 3;
				sp = alloc_state();
				if (!sp) goto error;
				*slp = sp;
				slp = &sp->next;
				sp->type = TYPE_SSL;
				goto fill_common;
			}
			goto fail;
		case 'G':
			if (!strncasecmp(cp, "gettime", 7)) {
				cp += 7;
				sp = alloc_state();
				if (!sp) goto error;
				*slp = sp;
				slp = &sp->next;
				sp->type = TYPE_GETTIME;
				goto fill_common;
			}
			if (!strncasecmp(cp, "gettofile", 9)) {
				cp += 9;
				type = TYPE_GETTOFILE;
				goto arg_fnm;
			}
			if (!strncasecmp(cp, "get", 3)) {
				cp += 3;
				type = TYPE_GET;
				goto arg_str;
			}
			goto fail;
		case 'F':
			if (!strncasecmp(cp, "failed", 6)) {
				cp += 6;
				flags |= FLAG_ONLY_FAILED;
				break;
			}
			goto fail;
		case 'P':
			if (!strncasecmp(cp, "proto", 5)) {
				cp += 5;
				flags |= FLAG_PRINT_PROTO;
				break;
			}
			goto fail;
		case 'C':
			if (!strncasecmp(cp, "casematch", 9)) {
				cp += 9;
				flags |= FLAG_CASEMATCH;
				break;
			}
			if (!strncasecmp(cp, "caseignore", 10)) {
				cp += 10;
				flags &= ~FLAG_CASEMATCH;
				break;
			}
			if (!strncasecmp(cp, "close", 5)) {
				cp += 5;
				sp = alloc_state();
				if (!sp) goto error;
				*slp = sp;
				slp = &sp->next;
				sp->type = TYPE_CLOSE;
				goto fill_common;
			}
			goto fail;
		case 'T':
			if (!strncasecmp(cp, "timeout", 7)) {
				cp += 7;
				sz = strspn(cp, " \t");
				cp += sz;
				if (!sz) goto fail;
				if (sscanf(cp, "%u%n", &u, &sz) != 1)
					goto fail;
				cp += sz;
				if (u > def_timeout) timeout = def_timeout;
				else timeout = u;
				break;
			}
			if (!strncasecmp(cp, "timeresolve", 11)) {
				cp += 11;
				flags |= FLAG_PRINT_TIME_RESOLVE;
				break;
			}
			if (!strncasecmp(cp, "timeconnect", 11)) {
				cp += 11;
				flags |= FLAG_PRINT_TIME_CONNECT;
				break;
			}
			goto fail;
		case 'E':
			if (!strncasecmp(cp, "escape", 6)) {
				cp += 6;
				sz = strspn(cp, " \t");
				cp += sz;
				if (!sz || *cp || '\n' == *cp || '\r' == *cp)
					goto fail;
				escape = *cp++;
				break;
			}
			goto fail;
		case 'H':
			if (!strncasecmp(cp, "holdssl", 7)) {
				cp += 7;
				flags |= FLAG_HOLD_SSL;
				break;
			}
		default:
			debug(3, "%s: Unknown command near: [%.10s]\n",
				__FUNCTION__, cp);
			goto fail;
		}

		to_next:
		sz = strspn(cp, " \t");
		cp += sz;
		if (!*cp || '\n' == *cp || '\r' == *cp) {
			if (states) {
				*states = slist;
				return 0;
			} else {
				/* Parse only */
				retval = 0;
				goto fail;
			}
		}
		if (!sz) goto fail;
		continue;

		fill_common:
		sp->timeout = timeout;
		sp->escape = escape;
		sp->flags = flags;

		/* Reset the following fields on each new state */
		timeout = 0;
		goto to_next;

		arg_fnm:
		sz = strspn(cp, " \t");
		cp += sz;
		if (!sz || !*cp) goto fail;
		if (!sp) {
			sp = alloc_state();
			if (!sp) goto error;
			*slp = sp;
			slp = &sp->next;
			sp->type = type;
		}
		sz = strcspn(cp, " \t\r\n");
		sp->fnm = (void *) malloc(sz + 1);
		if (!sp->fnm) goto error;
		memcpy(sp->fnm, cp, sz);
		sp->fnm[sz] = 0;
		cp += sz;
		if (TYPE_GETTOFILE == type) goto arg_str;
		goto fill_common;

		arg_str:
		sz = strspn(cp, " \t");
		cp += sz;
		if (!sz || !*cp) goto fail;
		if (!sp) {
			sp = alloc_state();
			if (!sp) goto error;
			*slp = sp;
			slp = &sp->next;
			sp->type = type;
		}
		sz4 = -1;
		c = 0;
		for (sz3 = 0; sz3 < 1000000;) {
			sz3 += 32;
			str = realloc(sp->str, sz3+1);
			if (!str) goto error;
			sp->str = str;
			if ('"' == *cp || '\'' == *cp) {
				sz2 = np_get_quoted_string(sp->str, sz3,
					&sz, cp, -1);
			} else {
				if (sz4 < 0) {
					sz4 = strcspn(cp, " \t\n\r");
					c = cp[sz4];
				}
				cp[sz4] = 0;
				sz2 = np_get_quoted_string(sp->str, sz3,
					&sz, cp, 0);
				cp[sz4] = c;
			}
			if (sz2 >= 0) break;
			if (-2 != sz2) goto fail;
		}
		if (!sz2 && (TYPE_GET == type || TYPE_GETTOFILE == type ||
			TYPE_SEND == type)) {
			/* Empty regex is not allowed for GET */
			debug(3, "%s: size=%d for command %s is not allowed.\n",
				__FUNCTION__, sz2, debug_cmd_str(type));
			goto fail;
		}
		sp->str_size = sz2;
		cp += sz;
		goto fill_common;
	}

error:
	retval = 3;
fail:
	if (slist) free_state(slist);
	return retval;
}

/*
** Parse a line from the input file in format:
** [options]host:port [parser_instructions]
** Return:
** <0: fatal/internal error
** 0: ok
** >0: user error
*/

static int parse_line(char *line, int *tcp_fd, int *udp_fd, int *sock)
{
char host[256], c, *ep;
unsigned port;
int n, s, sta, flags;
conn_t *cp;
state_t *states = 0;
long sid = 0;
int ttl = 0;

	*host = 0;
	port = 0;
	flags = 0;
	n = 0;
	s = 0;
	for (n = 0;;) {
		n += strspn(line+n, " \t");
		s = n;
		c = line[n] & 0xDF;
		if ('S' == c) {
			if (!strncasecmp(line+n, "SSL", 3)) {
				n += 3;
				n += strspn(line+n, " \t");
				if (',' != line[n++]) {
					/* May be the service is ssl:XXX */
					break;
				}
				flags |= CONN_SSL;
				continue;
			}
			if (!strncasecmp(line+n, "SID", 3)) {
				n += 3;
				n += strspn(line+n, " \t");
				if ('=' != line[n++]) {
					/* May be the service is sid:XXX */
					break;
				}
				n += strspn(line+n, " \t");
				errno = 0;
				ep = 0;
				sid = strtol(line+n, &ep, 0);
				if (!ep || (ep == line+n) || errno) {
					debug(1, "%s: Invalid sid in line: %s\n",
						__FUNCTION__, line);
					return 1;
				}
				n = ep - line;
				n += strspn(line+n, " \t");
				if (',' != line[n++]) {
					debug(1, "%s: Invalid data after sid in line: %s\n",
						__FUNCTION__, line);
					return 1;
				}
				continue;
			}
		}
		if ('U' == c) {
			if (!strncasecmp(line+n, "UDP", 3)) {
				n += 3;
				n += strspn(line+n, " \t");
				if (',' != line[n++]) {
					/* May be the service is udp:XXX */
					break;
				}
				flags |= CONN_UDP;
				continue;
			}
		}
		if ('T' == c) {
			if (!strncasecmp(line+n, "TTL", 3)) {
				n += 3;
				n += strspn(line+n, " \t");
				if ('=' != line[n++]) {
					/* May be the service is ttl:XXX */
					break;
				}
				n += strspn(line+n, " \t");
				errno = 0;
				ep = 0;
				ttl = strtol(line+n, &ep, 0);
				if (!ep || (ep == line+n) || errno) {
					debug(1, "%s: Invalid ttl in line: %s\n",
						__FUNCTION__, line);
					return 1;
				}
				n = ep - line;
				n += strspn(line+n, " \t");
				if (',' != line[n++]) {
					debug(1, "%s: Invalid data after ttl in line: %s\n",
						__FUNCTION__, line);
					return 1;
				}
				continue;
			}
		}
		break;
	}
	n = 0;
	sta = sscanf(line+s, " %255[^:,\n \t]%*[: \t]%u%n", host, &port, &n);
	if (2 != sta || !*host || !port || !n) {
		debug(1, "%s: Invalid service in line: %s\n", __FUNCTION__, line);
		return 1;
	}

	if (!with_ssl && flags & CONN_SSL) {
		debug(1, "%s: Ignoring SSL request %s:%u: SSL is not supported.\n",
			__FUNCTION__, host, port);
		return 1;
	}

	if (flags & CONN_SSL && flags & CONN_UDP) {
		debug(1, "%s: Ignoring request %s:%u: SSL with UDP is not supported.\n",
			__FUNCTION__, host, port);
		return 1;
	}

	if (flags & CONN_SSL) use_ssl = 1;

	n += s;
	s = strspn(line+n, " \t");
	n += s;

	s = parse_states(line+n, parse_only? 0 : &states);
	if (1 == s) {
		if (def_states)
			debug(3, "%s: Using default parser commands for %s:%u\n",
				__FUNCTION__, host, port);
		states = def_states;
		if (parse_only) return 0;
	} else if (2 == s || (!s && !sta)) {	/* 2=Invalid data */
		return 1;
	} else if (3 == s) {	/* Can't continue */
		return -1;
	}
	if (states) states->refcnt ++;

	cp = alloc_conn();
	if (!cp) {
		mem_error();
		if (states) free_state(states);
		return -1;
	}
	cp->sid = sid;
	cp->ttl = ttl;
	cp->states = states;
	cp->host = strdup(host);
	cp->flags = flags;
	if (!cp->host) {
		mem_error();
		free_conn(cp);
		return -1;
	}
	cp->port = port;
	if (parse_only) return 0;
	if (cp->flags & CONN_UDP) {
		*sock = *udp_fd;
		conns[*udp_fd] = cp;
		*udp_fd = -1;
	} else {
		*sock = *tcp_fd;
		conns[*tcp_fd] = cp;
		*tcp_fd = -1;
	}
	return 0;
}

static void print_result(int i)
{
int flags;
conn_t *cp;
state_t *sp;
char *spc;

	if (!conns[i]) return;
	cp = conns[i];
	flags = FLAG_PRINT_SERVICE | FLAG_PRINT_STATUS;
	if ((sp = cp->states) &&
		((flags = sp->flags) & FLAG_ONLY_FAILED) &&
		cp->status >= 0) goto end;
	spc = "";
	if (flags & FLAG_PRINT_PROTO) {
		fprintf(fpout, "%s%s", spc,
			(cp->flags & CONN_UDP)? "udp":
			((cp->flags & CONN_SSL)? "ssl":"tcp"));
		spc = " ";
	}
	if (flags & FLAG_PRINT_SERVICE) {
		fprintf(fpout, "%s%s:%u", spc, cp->host, cp->port);
		spc = " ";
	}
	if (flags & FLAG_PRINT_SID) {
		fprintf(fpout, "%s%ld", spc, cp->sid);
		spc = " ";
	}
	if (flags & FLAG_PRINT_STATUS) {
		fprintf(fpout, "%s%d", spc, cp->status);
		spc = " ";
	}
	if (flags & FLAG_PRINT_TIME_RESOLVE) {
		fprintf(fpout, "%s%d", spc, cp->time_resolve);
		spc = " ";
	}
	if (flags & FLAG_PRINT_TIME_CONNECT) {
		fprintf(fpout, "%s%d", spc, cp->time_connect);
		spc = " ";
	}
	if (cp->result) {
		fprintf(fpout, "%s%s", spc, cp->result);
	}
	if (*spc) fprintf(fpout, "\n");
end:
	free_conn(cp);
	conns[i] = 0;
}

static void print_results(void)
{
int i;

	for (i = min_sock; i <= max_sock; i++) print_result(i);
}

/*
** Return:
** -1: fatal error
** 0: checks completed
*/

static int parse_and_start_file_requests(void)
{
int sta, mf, have_entries, s, rc1, i, fl;
list_t *p, *l;
conn_t *cp;
struct sockaddr_in *adp;
struct timeval tv;
int tcp_fd = -1;
int udp_fd = -1;
int nsocks = socks;
int try = 0;

	/* OK, Now create sockets and manage connections	*/
	/* We maintain one free TCP and one free UDP socket	*/

	mf = 0;
	adp = 0;
	adp = (void *) malloc(sizeof(ad));
	have_entries = 1;
	ns = 0;
	FD_ZERO(&readfs);
	FD_ZERO(&writefs);

	/* Start timer */
	signal(SIGALRM, sig_alrm);
	alarm(1);

	for (;;) {

		while (have_entries && ns < nsocks) {
			if (tcp_fd < 0) {
				s = socket(AF_INET, SOCK_STREAM, 0);
				if (s >= 0) {
					tcp_fd = s;
					if (s < min_sock) min_sock = s;
					if (s > max_sock) max_sock = s;
					if (conns[s]) print_result(s);
				}
			}
			if (udp_fd < 0) {
				s = socket(AF_INET, SOCK_DGRAM, 0);
				if (s >= 0) {
					udp_fd = s;
					if (s < min_sock) min_sock = s;
					if (s > max_sock) max_sock = s;
					if (conns[s]) print_result(s);
				}
			}
			if (tcp_fd < 0 || udp_fd < 0) {
				debug(1, "socket(): %s\n", strerror(errno));
				if (ns <= 0) {
					/*
					** Wow, there is a BUG somewhere or
					** the system is very busy
					*/
					debug(1, "socket() fails when ns=%d\n",
						ns);
					return -1;
				}
				break;
			}
			if (!adp) {
				adp = (void *) malloc(sizeof(ad));
				if (!adp) {
					if (++mf > 100) {
						mem_error();
						goto label_exit;
					}
					continue;
				}
			}

			s = -1;
			for (;;) {
				if (have_one_request) {
					have_one_request = 0;
					strncpy(buffer,one_request,
						bufsize);
					buffer[bufsize] = 0;
				}
				else
				if (!have_file_requests ||
					!fgets(buffer,bufsize,fpin)) {
					have_entries = 0;
					break;
				}
				buffer[bufsize] = 0;
				sta = strspn(buffer, " \t");
				/* Comments */
				if ('#' == buffer[sta] ||
					';' == buffer[sta]) continue;
				if ('\n' == buffer[sta] ||
					'\r' == buffer[sta]) continue;
				/* parse_only=1: don't touch tcp_fd and udp_fd */
				s = -1;
				sta = parse_line(buffer+sta, &tcp_fd, &udp_fd, &s);
				if (!sta && !parse_only) break;
				if (sta < 0) return -1;
				s = -1;
			}
			if (s < 0) break;

			ns ++;

			try = 0;
			conns[s]->fd = s;
			timer_mod(&tt, &conns[s]->timer,
				conn_next_timeout(conns[s]));
			sta = start_connection(s);
			if (!sta) {
				FD_SET(s, &writefs);
				FD_SET(s, &readfs);
				state[s] = STATE_CONNECTING;
				sock_ad[s] = adp;
				adp = 0;
			} else if (sta > 0) {
				connect_finished(s);
				if (next_state(s)) {
					conns[s]->flags |= CONN_CAN_WRITE;
					if (handle_event(s) < 0) {
						return -1;
					}
				}
			} else {
				close_sock(s);
				conns[s]->status = sta;
			}
		}

		if (!ns && !have_entries) {
			return 0;
		}

		/* Now handle r/w	*/

		tv.tv_sec = 1;
		tv.tv_usec = 0;
		rfs = readfs;
		wfs = writefs;
		do {
			rc1 = select(FD_SETSIZE, &rfs, &wfs, 0, &tv);
		} while (rc1 < 0 && EINTR == errno);

		if ((rc1 < 0) && (ETIMEDOUT != errno)) {
			debug(3, "select(): %s\n", strerror(errno));
			goto before_continue;
		} else if (rc1 <= 0) {
			debug(5,"TIMEOUT, ns=%d\n", ns);

			if (++try >= def_timeout+5) {
				goto label_exit;
			}

			goto before_continue;
		}
		nready = rc1;

		debug(4, "nready=%d\n", nready);

		for (i = min_sock; (i <= max_sock) && (nready > 0); i++) {
			if (STATE_UNUSED == state[i]) {
				continue;
			}

			fl = 0;
			if (FD_ISSET(i,&wfs)) {
				fl ++;
				FD_CLR(i, &wfs);
				conns[i]->flags |= CONN_CAN_WRITE;
			}
			if (FD_ISSET(i,&rfs)) {
				fl ++;
				FD_CLR(i, &rfs);
				conns[i]->flags |= CONN_CAN_READ;
			}
			if (fl) {
				if (handle_event(i) < 0) {
					return -1;
				}
			}

			nready -= fl;

		}
#ifdef	NP_DEBUG
		if (np_debug_level >= 3) {
			if (nready > 0) {
				debug(3, "rest nready=%d\n", nready);
				//sleep(1);
			}
		}
#endif

		before_continue:

		if (flag_sig_alrm) {
			flag_sig_alrm = 0;
			/* Adjust the timer and expire connections */
			timer_adjust(&tt, 1);
			l = &tt.queue[tt.pos];
			for (p = l->n; p != l; ) {
				cp = list_entry(p, conn_t, timer);
				p = p->n;
				debug(4, "%s: Expiring %s, fd=%d\n",
					__FUNCTION__, debug_sid(cp),
					cp->fd);
				close_sock(cp->fd);
				/*
				** For UDP the final status is already set
				*/
				if (!(CONN_UDP & cp->flags)) {
					cp->status = -1;
				}
			}
			signal(SIGALRM, sig_alrm);
			alarm(1);
		}

	}

	label_exit:
	return -1;

}

int main(int argc,char *argv[])
{
int i;
struct sigaction act;
char *fnm, *ofnm;

	fnm = 0;
	ofnm = 0;
	for (i = 1; i < argc; i++) {
		if (('-' == *argv[i]) && (strcmp(argv[i],"-"))) {
			/* Number of sockets */
			if (!strcasecmp(argv[i],"-n")) {
				if (++i < argc) {
					socks = atoi(argv[i]);
					if ((socks <= 0) ||
						(socks > NUM_SOCKETS)) {
						debug(1, "Invalid value for %s: %s\n\n",
							argv[i-1], argv[i]);
						goto label_fail;
					}
				} else {
					goto label_fail;
				}
				continue;
			}
			/* Timeout */
			if (!strcasecmp(argv[i],"-to")) {
				if (++i < argc) {
					def_timeout = atoi(argv[i]);
					if ((def_timeout <= 0) ||
						(def_timeout > MAX_TIMEOUT)) {
						debug(1, "Invalid value for %s: %s\n\n",
							argv[i-1], argv[i]);
						goto label_fail;
					}
				} else {
					goto label_fail;
				}
				continue;
			}
			/* Max GET buffer size */
			if (!strcasecmp(argv[i],"-mg")) {
				if (++i < argc) {
					max_get_size = atoi(argv[i]);
					if ((max_get_size < 0) ||
						(max_get_size > MAX_GET_SIZE)) {
						debug(1, "Invalid value for %s: %s\n\n",
							argv[i-1], argv[i]);
						goto label_fail;
					}
				} else {
					goto label_fail;
				}
				continue;
			}
			/* res_retrans */
			if (!strcasecmp(argv[i],"-res_retrans")) {
				if (++i < argc) {
					res_retrans = atoi(argv[i]);
				} else {
					goto label_fail;
				}
				continue;
			}
			/* res_retry */
			if (!strcasecmp(argv[i],"-res_retry")) {
				if (++i < argc) {
					res_retry = atoi(argv[i]);
				} else {
					goto label_fail;
				}
				continue;
			}
			/* Service */
			if (!strcasecmp(argv[i],"-s")) {
				if (++i < argc) {
					if (have_one_request) {
						debug(1, "Too many %s options\n",
							argv[i-1]);
						goto label_fail;
					}
					have_one_request = 1;
					one_request = argv[i];
				} else {
					goto label_fail;
				}
				continue;
			}
			/* Input file */
			if (!strcasecmp(argv[i],"-f")) {
				if (++i < argc) {
					fnm = argv[i];
				} else {
					goto label_fail;
				}
				continue;
			}
			/* Output file */
			if (!strcasecmp(argv[i],"-o")) {
				if (++i < argc) {
					ofnm = argv[i];
				} else {
					goto label_fail;
				}
				continue;
			}
#ifdef	NP_DEBUG
			/* Debugging */
			if (!strcasecmp(argv[i],"-d")) {
				if (++i < argc) {
					np_debug_level = atoi(argv[i]);
					if (np_debug_level < 0) {
						fprintf(stderr, "Invalid value for %s: %s\n\n",
							argv[i-1], argv[i]);
						goto label_fail;
					}
				} else {
					goto label_fail;
				}
				continue;
			}
#endif
			/* Parse only */
			if (!strcasecmp(argv[i],"-p")) {
				parse_only = 1;
				continue;
			}
			goto label_fail;
		}
		if (def_command) {
			goto label_fail;
		}
		def_command = argv[i];
	}

	if (!fnm) {
		if (!have_one_request) {
			/* No -f and -s => assume stdin */
			have_file_requests = 1;
			fpin = stdin;
		}
	} else {
		if (strcmp(fnm, "-")) {
			fpin = fopen(fnm, "r");
			if (!fpin) {
				debug(1, "Unable to open infile %s\n", fnm);
				exit(1);
			}
		} else {
			fpin = stdin;
		}
		have_file_requests = 1;
	}

	fpout = stdout;
	if (ofnm && strcmp(ofnm, "-")) {
		fpout = fopen(ofnm, "w");
		if (!fpout) {
			debug(1, "Unable to create outfile %s\n", ofnm);
			exit(1);
		}
	}

	if (!socks) {
		socks = def_nsocks;
	}

	buffer = (char *) malloc(bufsize+1);
	if (!buffer) {
		mem_error();
		exit(1);
	}

	if (one_request && strlen(one_request) > bufsize) {
		debug(1, "The specified request is too long\n");
		exit(1);
	}

	if (def_command) {
		i = parse_states(def_command, parse_only? 0 : &def_states);
		if (3 == i) {
			mem_error();
			exit(1);
		}
		if (2 == i) {
			debug(1, "Invalid data in parser commands.\n");
			exit(1);
		}
		if (i) def_states = 0;
		if (def_states) def_states->refcnt ++;
	}

	act.sa_handler = SIG_IGN;
	act.sa_flags = SA_RESTART;
	sigemptyset(&act.sa_mask);
	sigaction(SIGPIPE,&act,0);

	if (res_retrans) _res.retrans = res_retrans;
	if (res_retry) _res.retry = res_retry;

	tt.rows = def_timeout + 2;
	timer_init(&tt);

	if (parse_and_start_file_requests() < 0) exit(1);
	print_results();
	fflush(fpout);
	exit(0);

	label_fail:
	res_init();
	fprintf(stderr,"NetParse %s by Julian Anastasov, check services in parallel\n\n", NETPARSE_VERSION);
#ifdef	NP_DEBUG
	fprintf(stderr,"Usage: [-n nsocks] [-s service] [-f infile] [-o outfile] [-p] [-to timeout] [-mg size] [-res_retrans n] [-res_retry n] [-d debug_level] [commands]\n");
#else
	fprintf(stderr,"Usage: [-n nsocks] [-s service] [-f infile] [-o outfile] [-p] [-to timeout] [-mg size] [-res_retrans n] [-res_retry n] [commands]\n");
#endif
	fprintf(stderr,"\n");
#ifdef	NP_DEBUG
	fprintf(stderr,"-d debug_level   Write to stderr: 0=quiet, def=%d\n", np_debug_level);
#endif
	fprintf(stderr,"-n nsocks        Number of parallel connection (max=%d), def=%d\n",
					NUM_SOCKETS, def_nsocks);
	fprintf(stderr,"-s service       Service to check\n");
	fprintf(stderr,"-f infile        Input file with requests, '-'=stdin\n");
	fprintf(stderr,"-o outfile       Use this output file instead of stdout\n");
	fprintf(stderr,"-to timeout      Timeout in seconds for each request, def=%d\n",
					def_timeout);
	fprintf(stderr,"-mg size         Max buffer size for GET, def=%d\n",
					max_get_size);
	fprintf(stderr,"-res_retrans n   Resolver Retransmission Time Interval, def=%d\n",_res.retrans);
	fprintf(stderr,"-res_retry n     Resolver Retry Count, def=%d\n",
					_res.retry);
	fprintf(stderr,"-p               Parse only, don't check\n");
	fprintf(stderr,"commands         Commands for the parser to use\n");
	fprintf(stderr,"\n");

	return 1;
}
